# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import subprocess
import math
import re
from Bio import SeqIO
from Bio.Align.Applications import MuscleCommandline
from Bio.Align.Applications import MafftCommandline
from io import StringIO

import fdog.libs.fasta as fasta_fn
import fdog.libs.output as output_fn

##### FUNCTIONS RELATED TO SEQ ALIGNMENT #####

def check_fasta36_executable(fdogPath):
    try:
        fasta36_cmd = '%s/bin/aligner/bin/ggsearch36' % fdogPath
        subprocess.check_output(fasta36_cmd, shell = True, stderr = subprocess.STDOUT)
        return('%s/bin/aligner/' % fdogPath)
    except subprocess.CalledProcessError as e:
        sys.exit('\033[91mERROR: FASTA36 at %s/bin/aligner/bin/ not executable!\033[0m' % fdogPath)


def do_align(aligner, fa_file):
    """ Do alignment using MUSCLE or MAFFT for a multiple fasta file
    Return a dictionary (SeqIO object) containing seq IDs and aligned sequences
    Note: if any input seq is longer than 12.000 aa/nt, only MAFFT can be used
    """
    if fasta_fn.check_long_seq(fa_file) == 1:
        aligner = 'mafft-linsi'
    if aligner == 'muscle':
        align_cline = MuscleCommandline(input = fa_file)
    else:
        align_cline = MafftCommandline(
            input = fa_file, localpair = True, maxiterate = 1000)
    try:
        stdout, stderr = align_cline()
        aln_io = StringIO(stdout)
        aln_seq = SeqIO.to_dict((SeqIO.parse(aln_io,'fasta')))
        return(aln_seq)
    except:
        sys.exit(
            'ERROR: Error doing alignment with %s for %s\n%s' % (aligner, fa_file, align_cline))


def calc_Kimura_dist(aln_dict, id_1, id_2, debug):
    """ Calculate Kimura distance for a pair of sequences
    Input is a dictionary of MSA (see do_align function).
    The Kimura distance is calculated based on perl module
    https://metacpan.org/pod/Bio::Align::ProteinStatistics#D-distance-methods
    """
    matches = 0
    total = 0
    if id_1 in aln_dict and id_2 in aln_dict:
        for a, b in zip(aln_dict[id_1].seq, aln_dict[id_2].seq):
            if a != '-' and b != '-':
                if a == b:
                    matches +=1
                total += 1
        if not total == 0:
            D = 1 - (matches/total)
        else:
            D = 1
        output_fn.print_debug(
            debug, 'Kimura distance',
            'kimura = round(- (math.log( 1 - %s - (0.2 * (%s ** 2)))), 5)' % (D, D))
        try:
            kimura = round(- (math.log( 1 - D - (0.2 * (D ** 2)))), 5)
        except:
            kimura = 999
        return(kimura)
    else:
        sys.exit('%s or %s not found in %s!' % (id_1, id_2, aln_dict))


def calc_aln_score(fa1, fa2, aln_strategy = 'local', debugCore = False):
    """ Calculate alignment score for genes in fa2 vs other genes in fa1
    Return dictionary {gene_id:aln_score}
    """
    fdog_path = os.path.realpath(__file__).replace('/libs/alignment.py','')
    fasta36_options = '%s %s -s BP62 -m 9 -d 0 -z -1 -E 100' % (fa1, fa2)
    if aln_strategy == 'global':
        fasta36_cmd = '%s/bin/aligner/bin/ggsearch36 %s' \
                                % (fdog_path, fasta36_options)
    elif aln_strategy == 'glocal':
        fasta36_cmd = '%s/bin/aligner/bin/glsearch36 %s' \
                                % (fdog_path, fasta36_options)
    else:
        fasta36_cmd = '%s/bin/aligner/bin/ssearch36 %s' \
                                % (fdog_path, fasta36_options)
    output_fn.print_debug(
        debugCore, 'ALN SCORE',
        'Calculate aln score using FASTA36: %s' % fasta36_cmd)
    try:
        fasta36_out = subprocess.run(
                [fasta36_cmd], shell = True, capture_output = True, check = True)
    except:
        sys.exit('Error running FASTA36\n%s' % fasta36_out)
    # returns score for genes in fa2
    aln_score = {}
    cand_dict = SeqIO.to_dict((SeqIO.parse(open(fa2), 'fasta')))
    for id in list(cand_dict.keys()):
        aln_score[id[0:60]] = 0
    results = fasta36_out.stdout.decode().split('\n')
    for l in results:
        if len(l) > 1:
            gene_id = l.split()[0]
            if gene_id in aln_score:
                if re.search('\(\s+\d+\)', l):
                    l = re.sub(r'\(\s+','(', l)
                aln_score[gene_id] = aln_score[gene_id] + int(l.split()[2])
    return(aln_score)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

from pysam import FastaFile
from Bio import SeqIO

import fdog.libs.zzz as general_fn


##### FUNCTIONS RELATED TO FASTA SEQ #####

def add_seq_to_dict(dict, id, seq):
    """ Add fasta sequence to a dictionary """
    if not id in dict:
        dict[id] = seq
    return(dict)


def read_fasta(fa_file):
    """ Read LARGE fasta file and return fasta object
    Sequence can be get using fasta_object.fetch(seq_id)
    """
    fasta_object = FastaFile(fa_file)
    return(fasta_object)


def write_fasta(fa_dict, out_file):
    """ Write sequences in SeqIO dict into output file """
    with open(out_file, 'w') as out:
        for seq in fa_dict:
            out.write('>%s\n%s\n' % (seq, fa_dict[seq].seq))


def append_to_fasta_file(fa_file, new_fa_dict):
    """ Append a dict of fasta seq to an existing fasta file """
    general_fn.check_file_exist(fa_file)
    existing_seq = SeqIO.to_dict(SeqIO.parse(open(fa_file),'fasta'))
    with open(fa_file, 'a') as fa_out:
        for id, seq in new_fa_dict.items():
            if not id in existing_seq:
                fa_out.write('>%s\n%s\n' % (id, seq))


def check_long_seq(fa_file):
    """ Check if any sequence longer than 12.000 aa/nt"""
    fa_seq = SeqIO.parse(open(fa_file),'fasta')
    for fa in fa_seq:
        if len(fa.seq) > 12000:
            return(1)
    return(0)


def remove_dup(fa_file):
    """ Remove duplicated sequences (filter by headers) """
    tmp = {}
    fa_seq = SeqIO.parse(open(fa_file),'fasta')
    for fa in fa_seq:
        if not fa.id in tmp:
            tmp[fa.id] = fa.seq
    with open(fa_file, 'w') as out:
        for id, seq in tmp.items():
            out.write('>%s\n%s\n' % (id, seq))



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import os
import sys
from Bio.Blast.Applications import NcbiblastpCommandline
import xml.etree.ElementTree as ET
import subprocess


##### FUNCTIONS RELATED TO BLAST #####

def do_blastsearch(
        query, blast_db, evalBlast = 0.00001, lowComplexityFilter = False):
    """ Perform blastp search for a query fasta file
    Return an XML string contains blast result
    """
    filter = 'no'
    if lowComplexityFilter == True:
        filter = 'yes'
    try:
        blastp_cline = NcbiblastpCommandline(
            query = query, db = blast_db, evalue = evalBlast, seg = filter,
            max_target_seqs = 10, outfmt = 5)
        stdout, stderr = blastp_cline()
        return(stdout)
    except:
        sys.exit(
            'ERROR: Error running blastp search for %s against %s'
            % (query, blast_db))


def parse_blast_xml(blast_output):
    """ Parse Blast XML output from a string variable
    Return a dictionary containing query ID, query length, together with a list
    of hits and their bit score, evalue, align len
    """
    blast_dict = {}
    root = ET.fromstring(blast_output)
    blast_dict['query'] = root[8][0][2].text
    blast_dict['query_len'] = root[8][0][3].text
    blast_dict['hits'] = {}
    for type_tag in root.findall(
            'BlastOutput_iterations/Iteration/Iteration_hits/Hit'):
        value = type_tag.findall('*')
        hit_id = 'NA'
        for i in type_tag.findall('*'):
            if i.tag == 'Hit_def':
                if not i.text in blast_dict['hits']:
                    hit_id = i.text
                    blast_dict['hits'][hit_id] = {}
            if i.tag == 'Hit_hsps':
                if hit_id in blast_dict['hits']:
                    blast_dict['hits'][hit_id]['bit_score'] = i[0][1].text
                    blast_dict['hits'][hit_id]['evalue'] = i[0][3].text
                    blast_dict['hits'][hit_id]['align_len'] = i[0][13].text
    return(blast_dict)


def make_blastdb(args):
    (specName, specFile, outPath, silent) = args
    blastCmd = 'makeblastdb -dbtype prot -in %s -out %s/coreTaxa_dir/%s/%s' % (specFile, outPath, specName, specName)
    if silent == True:
        blastCmd = blastCmd + '> /dev/null 2>&1'
    try:
        subprocess.call([blastCmd], shell = True)
    except:
        sys.exit('Problem with running %s' % blastCmd)
    fileInGenome = "../../searchTaxa_dir/%s/%s.fa" % (specName, specName)
    fileInBlast = "%s/coreTaxa_dir/%s/%s.fa" % (outPath, specName, specName)
    if not os.path.exists(fileInBlast):
        os.symlink(fileInGenome, fileInBlast)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to merge all output files (.extended.fa, .phyloprofile,
#  _forward.domains, _reverse.domains) in a given directory into one file each.
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: dosch@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from os import listdir as ldir
import argparse
import yaml
from pkg_resources import get_distribution

def createConfigPP(phyloprofile, domains_0, ex_fasta, directory, out):
    settings = dict(
        mainInput = '%s/%s' % (directory, phyloprofile),
        fastaInput = '%s/%s' % (directory, ex_fasta)
    )
    if not domains_0 == None:
        settings['domainInput'] = '%s/%s' % (directory, domains_0)
    settings['clusterProfile'] = 'TRUE'
    with open('%s.config.yml' % (out), 'w') as outfile:
        yaml.dump(settings, outfile, default_flow_style = False)


def main():
    version = get_distribution('fdog').version
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(version) + '.')
    parser.add_argument('-i', '--input',
                        help='Input directory, where all single output (.extended.fa, .phyloprofile, _forward.domains, _reverse.domains) can be found',
                        action='store', default='', required=True)
    parser.add_argument('-o', '--output', help='Output name', action='store', default='', required=True)
    args = parser.parse_args()

    directory = args.input
    out = args.output
    if not os.path.exists(os.path.abspath(directory)):
        sys.exit('%s not found' % directory)
    else:
        directory = os.path.abspath(directory)

    phyloprofile = None
    domains_0 = None
    domains_1 = None
    ex_fasta = None
    lines_seen = set()
    for infile in ldir(directory):
        if infile.endswith('.phyloprofile') and not infile == out + '.phyloprofile':
            if not phyloprofile:
                phyloprofile = out + '.phyloprofile'
                phyloprofile_out = open(phyloprofile, 'w')
            with open(directory + '/' + infile, 'r') as reader:
                lines = reader.readlines()
                for line in lines:
                    if line not in lines_seen: # not a duplicate
                        phyloprofile_out.write(line)
                        lines_seen.add(line)
        elif infile.endswith('_forward.domains') and not infile == out + '_forward.domains':
            if not domains_0:
                domains_0 = out + '_forward.domains'
                domains_0_out = open(domains_0, 'w')
            with open(directory + '/' + infile, 'r') as reader:
                lines = reader.readlines()
                for line in lines:
                    domains_0_out.write(line)
        elif infile.endswith('_reverse.domains') and not infile == out + '_reverse.domains':
            if not domains_1:
                domains_1 = open(out + '_reverse.domains', 'w')
            with open(directory + '/' + infile, 'r') as reader:
                lines = reader.readlines()
                for line in lines:
                    domains_1.write(line)
        elif infile.endswith('.extended.fa') and not infile == out + '.extended.fa':
            if not ex_fasta:
                ex_fasta = out + '.extended.fa'
                ex_fasta_out = open(ex_fasta, 'w')
            with open(directory + '/' + infile, 'r') as reader:
                lines = reader.readlines()
                for line in lines:
                    ex_fasta_out.write(line)
    if phyloprofile:
        phyloprofile_out.close()
    if domains_0:
        domains_0_out.close()
    if domains_1:
        domains_1.close()
    if ex_fasta:
        ex_fasta_out.close()

    createConfigPP(phyloprofile, domains_0, ex_fasta, directory, out)
    print('Done! Output files:\n%s/%s.*' % (directory,out))


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import subprocess
import pyhmmer


##### FUNCTIONS RELATED TO HMM #####

def create_hmm(aln_file, out_file):
    """ Create hmm file for an alinment file """
    hmmbuild_cmd = 'hmmbuild %s %s' % (out_file, aln_file)
    try:
        subprocess.run(
            [hmmbuild_cmd], shell = True,
            stdout = open(os.devnull, 'wb'), check = True)
    except:
        sys.exit('ERROR: Error running hmmbuild %s' % hmmbuild_cmd)


def do_hmmsearch(
        hmm_file, search_fa, evalHmmer = 0.00001, scoreCutoff = 10,
        hitLimit = 10, cpus = os.cpu_count()):
    """ Perform hmmsearch for a hmm file vs a multiple fasta file
    Return a dictionary of hits and their e-value and bit-score
    Only "top" hits are returned. The cutoff is defined by
    max_score / 100 * (100 - scoreCutoff)
    By default, only hits that have at least 90% of the best bit score
    are considers
    """
    hmm_hits = {}
    with pyhmmer.easel.SequenceFile(search_fa, digital = True, alphabet = pyhmmer.easel.Alphabet.amino()) as seq_file:
        sequences = list(seq_file)
    with pyhmmer.plan7.HMMFile(hmm_file) as hmm_file:
        try:
            for hits in pyhmmer.hmmsearch(
                    hmm_file, sequences, E = evalHmmer, cpus = cpus):
                if len(hits) > 0:
                    n = 0
                    for hit in hits:
                        if hit.score >= hits[0].score/100*(100-scoreCutoff):
                            if n < hitLimit:
                                hmm_hits[hit.name.decode('ASCII')] = (
                                    hit.evalue,hit.score)
                                n += 1
        except:
            sys.exit(
                'ERROR: Error running hmmsearch for %s agains %s'
                % (hmm_file, search_fa))
    return(hmm_hits)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to prepare data for fdog.
#  For each given genome FASTA file, It will create a folder within searchTaxa_dir
#  with the naming scheme of fdog ([Species acronym]@[NCBI ID]@[Proteome version]
#  e.g HUMAN@9606@3), a annotation file in JSON format in annotation_dir and
#  a blast DB in coreTaxa_dir folder (optional).
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import argparse
from pathlib import Path
from Bio import SeqIO
import multiprocessing as mp
from tqdm import tqdm
from ete3 import NCBITaxa
import re
import shutil
from datetime import datetime
import time
from pkg_resources import get_distribution
from collections import OrderedDict

import fdog.libs.zzz as general_fn
import fdog.libs.tree as tree_fn
import fdog.libs.addtaxon as add_taxon_fn


def parse_map_file(mapping_file, folIn):
    """ Create spec name from mapping file
    And also check if given input files in mapping file exist
    """
    name_dict = {}
    with open(mapping_file) as f:
      for line in f:
        if not '#' in line:
            tmp = line.split('\t')
            file_name = tmp[0]
            file_in = '%s/%s' % (folIn, file_name)
            general_fn.check_file_exist(file_in)
            tax_id = tmp[1].strip()
            try:
                tax_name = tmp[2].strip()
            except:
                tax_name = ''
            try:
                ver = tmp[3].strip()
            except:
                ver = datetime.today().strftime('%y%m%d')
            spec_name = add_taxon_fn.generate_spec_name(tax_id, tax_name, ver)
            name_dict[file_in] = spec_name
    return(name_dict)


def main():
    version = get_distribution('fdog').version
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(version) + '.')
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('-i', '--input', help='Path to input folder', action='store', default='', required=True)
    required.add_argument('-m', '--mapping',
                            help='Tab-delimited text file containing <fasta_file_name>tab<taxonID>tab<taxonName>tab<genome_version>. The last 2 columns are optional.',
                            action='store', default='', required=True)
    optional.add_argument('-o', '--outPath', help='Path to output directory', action='store', default='')
    optional.add_argument('-c', '--coreTaxa', help='Include this taxon to core taxa (i.e. taxa in coreTaxa_dir folder)', action='store_true', default=False)
    optional.add_argument('-a', '--noAnno', help='Do NOT annotate this taxon using fas.doAnno', action='store_true', default=False)
    optional.add_argument('--cpus', help='Number of CPUs used for annotation. Default = available cores - 1', action='store', default=0, type=int)
    optional.add_argument('--replace', help='Replace special characters in sequences by "X"', action='store_true', default=False)
    optional.add_argument('--delete', help='Delete special characters in sequences', action='store_true', default=False)
    optional.add_argument('--force', help='Force overwrite existing data', action='store_true', default=False)

    args = parser.parse_args()
    folIn = args.input
    folIn = os.path.abspath(folIn)
    mapping = args.mapping
    general_fn.check_file_exist(mapping)
    outPath = args.outPath
    if outPath == '':
        fdogPath = os.path.realpath(__file__).replace('/addTaxon.py','')
        pathconfigFile = fdogPath + '/bin/pathconfig.txt'
        if not os.path.exists(pathconfigFile):
            sys.exit('No pathconfig.txt found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
        with open(pathconfigFile) as f:
            outPath = f.readline().strip()
    outPath = os.path.abspath(outPath)
    noAnno = args.noAnno
    coreTaxa = args.coreTaxa
    cpus = args.cpus
    if cpus == 0:
        cpus = mp.cpu_count()-2
    replace = args.replace
    delete = args.delete
    add_taxon_fn.check_conflict_opts(replace, delete)
    force = args.force

    start = time.time()
    ### parse mapping file
    name_dict = parse_map_file(mapping, folIn)

    ### initiate paths
    Path(outPath + '/searchTaxa_dir').mkdir(parents = True, exist_ok = True)

    ### create file in searchTaxa_dir [and coreTaxa_dir]
    genome_jobs = []
    blast_jobs = []
    for f in name_dict:
        spec_name = name_dict[f]
        ## remove old folder if force is set
        if force:
            if os.path.exists(outPath + '/searchTaxa_dir/' + spec_name):
                shutil.rmtree(outPath + '/searchTaxa_dir/' + spec_name)
            if os.path.exists(outPath + '/coreTaxa_dir/' + spec_name):
                shutil.rmtree(outPath + '/coreTaxa_dir/' + spec_name)
        ## create jobs
        genome_path = '%s/searchTaxa_dir/%s' % (outPath, spec_name)
        Path(genome_path).mkdir(parents = True, exist_ok = True)
        genome_jobs.append([f, genome_path, spec_name, force, replace, delete])
        if coreTaxa:
            genome_file = '%s/%s.fa' % (genome_path, spec_name)
            blast_jobs.append([outPath, spec_name, genome_file, force, True])
    pool = mp.Pool(cpus)

    print('Parsing genome for %s species...' % len(genome_jobs))
    genome_out = []
    for _ in tqdm(pool.imap_unordered(add_taxon_fn.create_genome, genome_jobs),
            total=len(genome_jobs)):
        genome_out.append(_)
    out_msg = 'Output for %s can be found in %s within searchTaxa_dir'  % (spec_name, outPath)
    if len(blast_jobs) > 0:
        print('\nCreating Blast DB for %s species...' % len(blast_jobs))
        blast_out = []
        for _ in tqdm(pool.imap_unordered(add_taxon_fn.create_blastdb, blast_jobs),
                total=len(blast_jobs)):
            blast_out.append(_)
        out_msg = '%s, coreTaxa_dir' % out_msg

    ### create annotation
    if not noAnno:
        Path(outPath + '/annotation_dir').mkdir(parents = True, exist_ok = True)
        for f in name_dict:
            genome_file = '%s/searchTaxa_dir/%s/%s.fa' % (outPath, name_dict[f], name_dict[f])
            add_taxon_fn.create_annoFile(outPath, genome_file, cpus, force)
        if os.path.exists('%s/annotation_dir/tmp' % outPath):
            shutil.rmtree('%s/annotation_dir/tmp' % outPath)
        out_msg = '%s, annotation_dir' % out_msg

    end = time.time()
    print('==> Adding %s taxa finished in %s'  % (len(name_dict), '{:5.3f}s'.format(end - start)))
    print('==> %s' % out_msg)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import argparse
import subprocess
from ete3 import NCBITaxa
from pkg_resources import get_distribution
import time

import fdog.libs.preparation as prepare_fn
import fdog.libs.orthosearch as ortho_fn
import fdog.libs.corecompile as core_fn
import fdog.libs.fas as fas_fn
import fdog.libs.tree as tree_fn
import fdog.libs.output as output_fn


def main():
    version = get_distribution('fdog').version
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/fDOG/wiki")
    required = parser.add_argument_group('Required arguments')
    required.add_argument('--seqFile', help='Input file containing the seed sequence (protein only) in fasta format',
                            action='store', default='', required=True)
    required.add_argument('--jobName', help='Job name. This will also be file name for the output',
                            action='store', default='', required=True)
    required.add_argument('--refspec', help='Reference taxon. It should be the species the seed sequence was derived from',
                            action='store', default='', required=True)

    optional_paths = parser.add_argument_group('Non-default directory options')
    optional_paths.add_argument('--outpath', help='Output directory', action='store', default='')
    optional_paths.add_argument('--hmmpath', help='Path for the core ortholog directory', action='store', default='')
    optional_paths.add_argument('--corepath', help='Path for the core taxa directory', action='store', default='')
    optional_paths.add_argument('--searchpath', help='Path for the search taxa directory', action='store', default='')
    optional_paths.add_argument('--annopath', help='Path for the pre-calculated feature annotion directory', action='store', default='')
    optional_paths.add_argument('--pathFile', help='Config file contains paths to data folder (in yaml format)', action='store', default='')

    core_options = parser.add_argument_group('Core compilation options')
    core_options.add_argument('--coreOnly', help='Compile only the core orthologs', action='store_true', default=False)
    core_options.add_argument('--reuseCore', help='Reuse existing core set of your sequence', action='store_true', default=False)
    core_options.add_argument('--minDist', help='Minimum systematic distance of primer taxa for the core set compilation. Default: genus',
                            choices=['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'superkingdom'],
                            action='store', default='genus')
    core_options.add_argument('--maxDist', help='Maximum systematic distance of primer taxa for the core set compilation. Default: kingdom',
                            choices=['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'superkingdom'],
                            action='store', default='kingdom')
    core_options.add_argument('--coreSize', help='Maximul number of orthologs in core set. Default: 6', action='store', default=6, type=int)
    core_options.add_argument('--coreTaxa', help='List of primer taxa that should exclusively be used for the core set compilation', action='store', default='')
    core_options.add_argument('--CorecheckCoorthologsOff', help='Turn off checking for co-ortholog of the reverse search during the core compilation',
                                action='store_true', default=False)
    core_options.add_argument('--coreRep', help='Obtain only the sequence being most similar to the corresponding sequence in the core set rather than all putative co-orthologs',
                                action='store_true', default=False)
    core_options.add_argument('--coreHitLimit', help='Number of hits of the initial pHMM based search that should be evaluated via a reverse search. Default: 3',
                                action='store', default=3, type=int)
    core_options.add_argument('--distDeviation', help='The deviation in score in percent (0 = 0 percent, 1 = 100 percent) allowed for two taxa to be considered similar. Default: 0.05',
                                action='store', default=0.05, type=float)
    core_options.add_argument('--alnStrategy', help='Specify the alignment strategy during core ortholog compilation. Default: local',
                                choices=['local', 'glocal', 'global'],
                                action='store', default='local')

    ortho_options = parser.add_argument_group('Ortholog search strategy options')
    ortho_options.add_argument('--searchTaxa', help='Specify file contains list of search taxa', action='store', default='')
    ortho_options.add_argument('--group', help='Allows to limit the search to a certain systematic group', action='store', default='')
    ortho_options.add_argument('--checkCoorthologsRefOff', help='Turn off checking for co-ortholog of the reverse search during the final ortholog search',
                                action='store_true', default=False)
    ortho_options.add_argument('--rbh', help='Requires a reciprocal best hit during the ortholog search to accept a new ortholog',
                                action='store_true', default=False)
    ortho_options.add_argument('--rep', help='Obtain only the sequence being most similar to the corresponding sequence in the core set rather than all putative co-orthologs',
                                action='store_true', default=False)
    ortho_options.add_argument('--lowComplexityFilter', help='Switch the low complexity filter for the blast search on. Default: False',
                                action='store_true', default=False)
    ortho_options.add_argument('--evalBlast', help='E-value cut-off for the Blast search. Default: 0.00001',
                                action='store', default=0.00005, type=float)
    ortho_options.add_argument('--evalHmmer', help='E-value cut-off for the HMM search. Default: 0.00001',
                                action='store', default=0.00005, type=float)
    ortho_options.add_argument('--hitLimit', help='number of hits of the initial pHMM based search that should be evaluated via a reverse search. Default: 10',
                                action='store', default=10, type=int)
    ortho_options.add_argument('--scoreCutoff', help='Define the percent range of the hmms core of the best hit up to which a candidate of the hmmsearch will be subjected for further evaluation. Default: 10',
                                action='store', default=10, type=int)

    fas_options = parser.add_argument_group('FAS options')
    fas_options.add_argument('--coreFilter',
                                help='Specifiy mode for filtering core orthologs by FAS score. In \'relaxed\' mode candidates with insufficient FAS score will be disadvantaged. In \'strict\' mode candidates with insufficient FAS score will be deleted from the candidates list. The option \'--minScore\' specifies the cut-off of the FAS score.',
                                choices=['relaxed', 'strict'], action='store', default='')
    fas_options.add_argument('--minScore', help='Specify the threshold for coreFilter. Default: 0.75', action='store', default=0.75, type=float)

    addtionalIO = parser.add_argument_group('Other I/O options')
    addtionalIO.add_argument('--append', help='Append the output to existing output files', action='store_true', default=False)
    addtionalIO.add_argument('--force', help='Overwrite existing ortholog search output files', action='store_true', default=False)
    addtionalIO.add_argument('--forceCore', help='Overwrite existing core set of your sequence', action='store_true', default=False)
    addtionalIO.add_argument('--noCleanup', help='Temporary output will NOT be deleted. Default: False', action='store_true', default=False)
    addtionalIO.add_argument('--debug', help='Set this flag to obtain more detailed information about the ortholog search progress', action='store_true', default=False)
    addtionalIO.add_argument('--debugCore', help='Set this flag to obtain more detailed information about the core compilation actions', action='store_true', default=False)
    addtionalIO.add_argument('--silentOff', help='Show more output to terminal', action='store_true', default=False)

    optional = parser.add_argument_group('Other options')
    optional.add_argument('--fasOff', help='Turn OFF FAS support', action='store_true', default=False)
    optional.add_argument('--aligner', help='Choose between mafft-linsi or muscle for the multiple sequence alignment. DEFAULT: muscle',
        choices=['mafft-linsi', 'muscle'], action='store', default='muscle')
    optional.add_argument('--cpus', help='Determine the number of threads to be run in parallel. Default: 4', action='store', default=4, type=int)

    ### get arguments
    args = parser.parse_args()

    # required arguments
    seqFile = args.seqFile
    seqName = args.jobName
    refspec = args.refspec

    # path arguments
    outpath = os.path.abspath(args.outpath)
    hmmpath = args.hmmpath
    corepath = args.corepath
    searchpath = args.searchpath
    annopath = args.annopath
    pathFile = args.pathFile

    # core compilation arguments
    coreOnly = args.coreOnly
    reuseCore = args.reuseCore
    minDist = args.minDist
    maxDist = args.maxDist
    coreSize = args.coreSize
    coreTaxa = args.coreTaxa
    if not coreTaxa == '':
        if os.path.exists(os.path.abspath(coreTaxa)):
            coreTaxa = os.path.abspath(coreTaxa)
    CorecheckCoorthologsOff = args.CorecheckCoorthologsOff
    coreRep = args.coreRep
    coreHitLimit = args.coreHitLimit
    distDeviation = args.distDeviation
    alnStrategy = args.alnStrategy

    # ortholog search arguments
    searchTaxa = args.searchTaxa
    if not searchTaxa == '':
        if os.path.exists(os.path.abspath(searchTaxa)):
            searchTaxa = os.path.abspath(searchTaxa)
    group = args.group
    if not group == '' and not searchTaxa == '':
        print('WARNING: Both --group and --searchTaxa are specified. Search taxa will be obtained only from %s!' % searchTaxa)
        group = ''
    checkCoorthologsRefOff = args.checkCoorthologsRefOff
    rbh = args.rbh
    rep = args.rep
    lowComplexityFilter = args.lowComplexityFilter
    evalBlast = args.evalBlast
    evalHmmer = args.evalHmmer
    hitLimit = args.hitLimit
    scoreCutoff = args.scoreCutoff

    # fas arguments
    fasOff = args.fasOff
    coreFilter = args.coreFilter
    minScore = args.minScore

    # other I/O arguments
    append = args.append
    force = args.force
    forceCore = args.forceCore
    noCleanup = args.noCleanup
    debug = args.debug
    debugCore = args.debugCore
    silentOff = args.silentOff

    # others
    aligner = args.aligner
    cpus = args.cpus
    if cpus > os.cpu_count():
        cpus = os.cpu_count()


    begin = time.time()
    ##### Check and group parameters
    if seqFile == 'infile.fa':
        fdogPath = os.path.realpath(__file__).replace('/runSingle.py','')
        seqFile = '%s/data/infile.fa' % fdogPath

    (seqFile, hmmpath, corepath, searchpath, annopath) = prepare_fn.check_input(
                    [seqFile, refspec, outpath, hmmpath,
                    corepath, searchpath, annopath, pathFile])
    pathArgs = [outpath, hmmpath, corepath, searchpath, annopath]

    if not fasOff:
        check_fas = fas_fn.check_fas_executable()
        if check_fas == 0:
            sys.exit('ERROR: FAS is not executable! You still can use fDOG with --fasOff!')

    ##### Identify seed ID from refspec genome
    seed_id = prepare_fn.identify_seed_id(seqFile, refspec, corepath, debug, silentOff)
    print('Identified seed ID: %s' % seed_id)


    ##### DO CORE COMPILATION
    # start = time.time()
    coreArgs = [minDist, maxDist, coreSize, coreTaxa, distDeviation,
                alnStrategy, fasOff]
    orthoCoreArgs = [CorecheckCoorthologsOff, rbh, True, evalBlast,
                    lowComplexityFilter, evalHmmer, coreHitLimit,
                    scoreCutoff, aligner] # rep = True
    otherCoreArgs = [cpus, debugCore, silentOff, noCleanup, force, append]
    print('Compiling core set for %s' % seqName)
    core_runtime = core_fn.run_compile_core([seqFile, seqName, refspec, seed_id, reuseCore,
        forceCore, coreArgs, pathArgs, orthoCoreArgs, otherCoreArgs, debug])
    # end = time.time()
    print('==> Core compilation finished in %s' % core_runtime[1])


    ##### DO ORTHOLOG SEARCH USING CORE HMM (HAMSTR)
    if not coreOnly:
        start = time.time()
        # check existing output
        finalOutfile = '%s/%s.extended.fa' % (outpath, seqName)
        finalOutfile = os.path.abspath(finalOutfile)
        output_fn.check_output_exist(finalOutfile, force, append)
        # get list of search taxa
        if not group == '':
            ### Check valid taxonomy group
            ncbi = NCBITaxa()
            group_id = ncbi.get_name_translator([group])
            if len(group_id) == 0:
                exit('ERROR: Taxon group "%s" invalid!' % group)
            ### create taxonomy tree from list of search taxa
            searchTaxa = []
            tax_ids = core_fn.get_core_taxa_ids(coreTaxa, corepath)

            for tax_id in tax_ids.keys():
                check = tree_fn.check_taxon_group(group_id[group][0], tax_id, ncbi)
                if check == True:
                    searchTaxa.append(tax_ids[tax_id])
            if debugCore:
                print(searchTaxa)
            if len(searchTaxa) == 0:
                exit('ERROR: No taxon found within %s taxonomy group!' % group)
            else:
                searchTaxa = ','.join(searchTaxa)
        # do ortholog search
        orthoArgs = [checkCoorthologsRefOff, rbh, rep, evalBlast,
                    lowComplexityFilter, evalHmmer, hitLimit, scoreCutoff, aligner]
        otherArgs = [searchTaxa, cpus, debug, silentOff, noCleanup, force, append]
        hamstr_out = ortho_fn.run_hamstr([seqName, refspec, pathArgs, orthoArgs, otherArgs])
        output_fn.write_hamstr(hamstr_out, outpath, seqName, force, append)
        end = time.time()
        print('==> Ortholog search finished in ' + '{:5.3f}s'.format(end - start))

        ##### DO FINAL FAS CALCULATION
        if not fasOff:
            try:
                fasVersion = subprocess.run(['fas.run --version'], shell = True, capture_output = True, check = True)
            except:
                sys.exit('Problem with FAS! Please check https://github.com/BIONF/FAS or turn it off if not needed!')
            if os.path.exists(finalOutfile):
                start = time.time()
                fas_fn.calc_fas_multi(finalOutfile, outpath, annopath, cpus)
                end = time.time()
                print('==> FAS calculation finished in ' + '{:5.3f}s'.format(end - start))
        else:
            output_fn.hamstr_2_profile(finalOutfile)

        end = time.time()
        print('==> fdog.run finished in ' + '{:5.3f}s'.format(end - begin))

if __name__ == '__main__':
    main()
